#!/usr/bin/python

'''
The purpose of this code is to demonstrate estimators for clique size distribution using egocentric sampled data.
It is released as a part of the paper:

Estimating Clique Composition and Size Distributions from Sampled Network Data
Minas Gjoka, Emily Smith, Carter Butts
UC Irvine


Changelog
****************
v1: July 2013
Initial release


'''

__author__ = """Minas Gjoka"""

import sys
import cPickle as pickle

class CliqueEstimation:

  def __init__(self):
    self.C = None
    self.Est = None

  def load_calculation(self, clique_calculation_fname):
    try:
      self.C = pickle.load(open(clique_calculation_fname))
      print "File '%s' successfully loaded" % (clique_calculation_fname)
      
      required_keys = ['n', 'N', 'replacement', 'sample_type', 'nsamples', 'weight', \
                       'labeling', 'attributes', ]
      
      for key in required_keys:
        if key not in self.C:
          print  "Invalid input file '%s'. Key '%s' not found." % (clique_calculation_fname, key)
          self.C = None

      if 'cliques'                  not in self.C and \
         'cliques_label'            not in self.C and \
         'cliques_attributes'       not in self.C and \
         'cliques_attributes_label' not in self.C:
        print  "Invalid input file '%s'. Clique calculation not found." % (clique_calculation_fname)
        self.C = None
      
    except:
      print "Invalid input file '%s' " % (clique_calculation_fname)

      
  def _prob_exclusion_wout_replacement(self,N,n,order):
    # N is graph size
    # n is sample size
    # order is clique size
    res1 = np.e**sum([ np.log( 1.*(N-n-i)/(N-i) )  for i in range(0,order)]  )
    return res1
      
  def estimate(self):
    if self.C == None:
      print "Unable to estimate. Clique calculation of egonet samples has not been loaded."
      return 

    n = self.C['n']   # sample size
    N = self.C['N']   # population size
    replacement = self.C['replacement'] # 'with' or 'wout'
    sample_type = self.C['sample_type'] # 'uis' or 'wis'
    nsamples = self.C['nsamples']       # number of times each node is sampled 
    weight = self.C['weight']           # sampling weight for each node
    
    labeling = self.C['labeling']  # 0 or 1
    has_attributes = self.C['attributes'] # 0 or 1
              
    self.Est = {}
    if sample_type == 'uis':
      
      if  (not labeling) and (not has_attributes):
        calculation_name = 'cliques'              
    
        D_i = {}
        for node,egocliq in self.C[calculation_name].iteritems():
          for idx,order in enumerate(egocliq):
            D_i[idx] = D_i.get(idx,0) + nsamples[node]*order
  
        for order,value in D_i.iteritems():
          if order!=0 and value!=0:
            self.Est[order] = (1.*value*N)/(1.*order*n)

            
      elif (labeling) and (not has_attributes):
        calculation_name = 'cliques_label'      
        
        uniq_cliq = set()
        DUniq_i = {}
        for node,egocliq_label in self.C[calculation_name].iteritems():
          for order, hash_cliqid in egocliq_label:   
            if hash_cliqid not in uniq_cliq:
              uniq_cliq.add(hash_cliqid)
  
              DUniq_i[order] = DUniq_i.get(order,0) + 1
              
        
              
       
        for order,value in DUniq_i.items():
          if order!=0 and value!=0:
            if   replacement=="with":
              inclusion_prob = 1 - (1 - float(order)/N)**n
            elif replacement=="wout":
              inclusion_prob = 1 - self._prob_exclusion_wout_replacement(N,n,order)
  
            self.Est[order] = 1.*value/inclusion_prob

      elif (not labeling) and (has_attributes):
        calculation_name = 'cliques_attributes'    
        
        D_u = {}
        for node,egocliq_attr in self.C[calculation_name].iteritems():
          for zero in egocliq_attr:
            for one in egocliq_attr[zero]:
              D_u.setdefault(zero,{})[one] = D_u.setdefault(zero,{}).get(one,0) + nsamples[node]*egocliq_attr[zero][one]
            
        for zero in D_u:
          for one in D_u[zero]:
            order = zero+one
            self.Est.setdefault(zero,{})[one] = (1.*D_u[zero][one]*N)/(1.*order*n)

      elif (labeling) and (has_attributes):
        calculation_name = 'cliques_attributes_label'
        
        uniq_cliq = set()
        DUniq_u = {}
        for node, egocliq_attr_label in self.C[calculation_name].iteritems():
          for zero in egocliq_attr_label:
            for one in egocliq_attr_label[zero]:                
              for hash_cliqid in egocliq_attr_label[zero][one]:
                if hash_cliqid not in uniq_cliq:
                  uniq_cliq.add(hash_cliqid)

                  DUniq_u.setdefault(zero,{})[one] = DUniq_u.setdefault(zero,{}).get(one,0) + 1
                          
        for zero in DUniq_u:
          for one,value in DUniq_u[zero].iteritems():
            order = zero + one
            if (order)!=0 and value!=0:
              if   replacement=="with":
                inclusion_prob = 1 - (1 - float(order)/N)**n
              elif replacement=="wout":
                inclusion_prob = 1 - self._prob_exclusion_wout_replacement(N,n,order)  

              self.Est.setdefault(zero,{})[one] = value/inclusion_prob
              
      elif sample_type == 'wis':
        print "This release does not support 'wis' sampling method"
        self.Est = None
      else:
        print "Unsupported sampling method"
        self.Est = None
  
  
  def dump(self,output_fname):
    if self.Est == None:
      print "Unable to dump estimation result because it is unavailable. "
      return
    
    if output_fname[ len(output_fname)-7:] != ".pickle":
      output_fname = '%s.pickle' % output_fname
    f = open(output_fname,"w")
    pickle.dump(self.Est,f,protocol=2)
    f.close()
    print "\nEstimation result dumped at file '%s'" % (output_fname)
    print "Press Enter to print estimation result"
    raw_input()
    print self.Est

      


if __name__ == "__main__":
  try:
    clique_calculation_fname = sys.argv[1]
    output_fname = sys.argv[2]
  except:
    print "Usage: %s clique_calculation.pickle  output" % (sys.argv[0])
    print "clique_calculation.pickle -> filename of pickle with sampled egonets and calculated clique distributions (run script SampleEgonets_CalculateCliques)"
    print "output -> filename to write output"
    print ""
    print "Script 'SampleEgonets_CalculateCliques' demonstrates how to prepare a 'clique_calculation.pickle' instance"
    sys.exit(0)


  print "\n"
  print "Author     : Minas Gjoka"
  print "Name       : EstimateCliques"
  print "Description: Demonstration of Estimators for Clique Size Distributions"
  #print "Tested with Python versions 2.6.x and 2.7.x"
  print "-------------------------------------------"
  print "Press 'Enter' to start"
  raw_input()

  try:
    import networkx as nx
    import numpy as np
  except:
    print "Error encountered. Requires Python packages 'networkx', 'numpy'"
    sys.exit()



  ##1.
  ce = CliqueEstimation()
  ce.load_calculation(clique_calculation_fname)
  ce.estimate()           # apply clique distribution estimators
  ce.dump(output_fname)
