#!/usr/bin/python

'''
The purpose of this code is to create egonet sample files  that can be used  as an input for clique estimation.
It is released as a part of the paper:

Estimating Clique Composition and Size Distributions from Sampled Network Data
Minas Gjoka, Emily Smith, Carter Butts
UC Irvine


Changelog
****************
v1: July 2013
Initial release


'''

__author__ = """Minas Gjoka"""



import sys, random, math
import sampling3 as sampling
import cPickle as pickle



class EgonetSampling:

  def __init__(self):
        
    self.g = None
    self.has_attributes = None
    self.node_attributes = None
    self.nsamples = None
    self.weight = None
    self.egonet_list = None

  def countOccur(self,input_list):
    h = {}
    for elem in input_list:
      h[elem] = h.get(elem,0) + 1.
    return h

  def load_graph(self, graph_name, attribute_name):
    self.graph_name = graph_name
    self.attribute_name = attribute_name
    print "Loading graph %s.." % (self.graph_name) ,
    sys.stdout.flush()

    if graph_name=="test":
      self.g = nx.generators.powerlaw_cluster_graph(1000,15,1)
      print "Done"            
    else:
      try:
        self.g = nx.read_edgelist(graph_name)        
        print "Done"
      except:
        print "\nUnable to load graph file. Is file in 'edge list format'?"
        return
      
      
    if self.attribute_name.lower() == 'none':
      self.has_attributes = 0
    else:      
      try:
        fin = open(attribute_name)        
        try:        
          self.node_attributes = dict( (int(key),int(val)) for (key,val) in ( line.split(":") for line in map(str.strip, open(attribute_name).readlines())  ) )
          if len(set(self.node_attributes.values()))>2:
            print "This release only supports binary node attributes whereas input file has %d attributes.\nNo attributes used." % ( len(set(node_attributes.values())) )
            self.has_attributes = 0
          elif len( set(int(i) for i in self.g.nodes()) - set(self.node_attributes) )>0:
            print "%d nodes do not have attributes.\nNo attributes used." % (len( set(int(i) for i in self.g.nodes()) - set(self.node_attributes) ))
          else:
            self.has_attributes = 1
        except:
          print "Error opening attributes file.\nNo attributes used"
          self.has_attributes = 0        
      except:
        print "Attributes file '%s' does not exist.\nNo attributes used" % attribute_name
        self.has_attributes = 0
      

  def sample(self, n, sample_type, replacement):
    if self.g == None:
      print "Unable to sample. Graph not loaded." 
      return
    
    self.n = n
    self.sample_type = sample_type
    self.replacement = replacement
    
    
    if not self.g:
        print "No graph loaded."
        return
    else:
        print "[%s] %d egonet samples " % (sample_type, self.n)    
    
    if self.sample_type=='uis':

      if self.replacement=="wout":
        sample_list = random.sample(self.g.nodes(),n)
        self.nsamples = dict([ (sample,1) for sample in set(sample_list)])            

      elif replacement=="with":

        sample_list = list(sampling.uniform_independent_node_sample(self.g, size=n))          
        self.nsamples = self.countOccur(sample_list)          

      else:
        print "**Unsupported replacement method '%s'**" % (replacement)
        return

      self.weight = dict( (node,1) for node in self.nsamples)  # 'uis' sampling
      
    elif self.sample_type=='wis':
      print "**This release does not support 'wis' sampling method**"
      return
    else:
      print "**Unsupported sampling method '%s'**" % (self.sample_type)
      return
      
    self.egonet_list = {}
    for node in self.nsamples:
      self.egonet_list[node] = nx.ego_graph(self.g, node)
    
    
class EgonetCliqueCalculation:
  def __init__(self, egonet_sample):    
    self.egonet_sample = egonet_sample
    self.Calc = None  # store results of exact clique calculation for each egonet
  
  def findcliques_binaryattr(self, g, v_attr):
    ### assume binary attributes of 0 and 1
    h = {}
    for cliq in nx.find_cliques(g):
      one = sum(v_attr[v] for v in cliq)
      zero = len(cliq) - one
      h.setdefault(zero,{})[one] = h.setdefault(zero,{}).get(one,0) + 1
    return h
  
  def findcliques_binaryattr_label(self, g, v_attr):    
    ### assume binary attributes of 0 and 1
    h = {}  
    for cliq in nx.find_cliques(g):
      one = sum(v_attr[v] for v in cliq)
      zero = len(cliq) - one
      h.setdefault(zero,{}).setdefault(one,[]).append(  hash('_'.join(str(v) for v in sorted(cliq)) ) )
    return h
    
  
  def calculate(self, labeling):
    if self.egonet_sample.egonet_list == None:
      print "Unable to calculate cliques. Egonets have not been sampled."
      return
    
    self.labeling = labeling
    has_attributes = self.egonet_sample.has_attributes

    calculation_name = ""
    if  (not labeling) and (not has_attributes):
      calculation_name = 'cliques'      
    elif (labeling) and (not has_attributes):
      calculation_name = 'cliques_label'      
    elif (not labeling) and (has_attributes):
      calculation_name = 'cliques_attributes'      
    elif (labeling) and (has_attributes):
      calculation_name = 'cliques_attributes_label'
      
    self.Calc = {} 
    self.Calc[calculation_name] = {}  
    
    maxcliq_uniqid = lambda cliqlist: ( len(cliqlist), hash('_'.join(str(v) for v in sorted(cliqlist)) ) )
    
    dot_freq = int( math.ceil(.01* len(self.egonet_sample.egonet_list) ))
    for idx, (ego, g_ego) in enumerate(self.egonet_sample.egonet_list.iteritems()):
      
      if   (not labeling) and (not has_attributes):
        self.Calc['cliques'][ego] = np.bincount( map(len,nx.clique.find_cliques(g_ego)) ) 
        
      elif (labeling) and (not has_attributes):
        self.Calc['cliques_label'][ego] = map(maxcliq_uniqid,nx.clique.find_cliques(g_ego))
        
      elif (not labeling) and (has_attributes):
        self.Calc['cliques_attributes'][ego] =  self.findcliques_binaryattr(g_ego, self.egonet_sample.node_attributes)  
        
      elif (labeling) and (has_attributes):
        self.Calc['cliques_attributes_label'][ego] = self.findcliques_binaryattr_label(g_ego, self.egonet_sample.node_attributes)
        
      else:
        print "All cases covered. This case is impossible."
        
      if idx % dot_freq == 0:
        sys.stdout.write(".")
        sys.stdout.flush()  
    
  
class CliqueCalculationStorage:
    def __init__(self, egonet_sample, egonet_clique_calculation):
      self.egonet_sample = egonet_sample
      self.egonet_clique_calculation = egonet_clique_calculation
      
    def dump(self, fname_write):
      if self.egonet_clique_calculation.Calc == None:
        print "Unable to dump. Egonet cliques have not been calculated."
        return 
      
      self.C = {
        'n':           self.egonet_sample.n,
        'sample_type': self.egonet_sample.sample_type ,
        'replacement': self.egonet_sample.replacement,
        'labeling':    self.egonet_clique_calculation.labeling,
                
        'N':          self.egonet_sample.g.number_of_nodes(),
        'attributes': self.egonet_sample.has_attributes,
        'nsamples':   self.egonet_sample.nsamples,
        'weight':     self.egonet_sample.weight
      }
      for key in self.egonet_clique_calculation.Calc:
        self.C[key] = self.egonet_clique_calculation.Calc[key]

      print "\nn:%d N:%d sample_type:%s replacement:%s" % (self.C['n'],self.C['N'],self.C['sample_type'],self.C['replacement'])
      print 'labeling:%d attributes:%d' % (self.C['labeling'],self.C['attributes'])
      print "\nStoring clique calculation for sampled egonets in file '%s'" % (fname_write)
      f = open(fname_write,"w")
      pickle.dump(self.C,f,protocol=2)
      f.close()
      print "The file '%s' can be used as input to the 'EstimateCliques' script" % (fname_write)
        
if __name__ == "__main__":
  try:
    graph_name = sys.argv[1]
    n = int(sys.argv[2])
    sample_type = sys.argv[3].lower()
    replacement = sys.argv[4].lower()
  
    labeling = int(sys.argv[5])
    attribute_name = sys.argv[6]

  except:
    print "Usage: %s graph_name sample_size sample_type replacement labeling attribute_name" % (sys.argv[0])
    print "graph_name -> filename of input graph in edgelist format"
    print "sample_size -> egonet sample size"
    print "sample_type -> 'uis' or 'wis'"
    print "replacement -> 'with' or 'wout'"
    print "labeling -> '0' or '1'"
    print "attribute_name -> 'none' or filename of attributes file for each node in input graph.\nthis code supports only binary attributes."
    print ""
    print "Example:"
    print "%s test 50 uis wout 0 none " % (sys.argv[0])
    print "(test generates a powerlaw_cluster graph using networkx)"
    sys.exit(0)
    
    
  print "\n"
  print "Author     : Minas Gjoka"
  print "Name       : SampleEgonets_CalculateCliques"
  print "Description: This script demonstrates how to prepare the data for clique estimation. More specifically, it samples egonets from a known graph and calculates the clique distribution for each sampled egonet"
  #print "Tested with Python versions 2.6.x and 2.7.x"
  print "-------------------------------------------"
  print "Press 'Enter' to start"
  raw_input()

  try:
    import networkx as nx
    import numpy as np
  except:
    print "Error encountered. Requires Python packages 'networkx', 'numpy'"
    sys.exit()
  
  

  ##1.
  es = EgonetSampling()
  es.load_graph(graph_name, attribute_name)  # loads graph 
  es.sample(n, sample_type, replacement)     # samples 'n' egonets from graph
    
  ##2.
  ecc = EgonetCliqueCalculation(es)  
  ecc.calculate(labeling) # calculates clique distribution for each egonet in the sample
  
  ##3.
  ccs = CliqueCalculationStorage(es, ecc)
  fname_write = '%s_sample_calculation.pickle' % graph_name
  ccs.dump(fname_write)
  